package org.jboss.jms.client.container;

/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005-2009, Red Hat Middleware LLC, and individual contributors
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;

/**
 * A ThreadValveCounter
 * 
 * This is a utility class to avoid a calling thread waiting on itself. In cases where an object is accessed concurrently,
 * closing that object requires first waiting for all the current ongoing calls to be finished, which means all current calling 
 * threads (except the one who is calling close()) returning from their respective method calls on that object. Sometimes
 * the calling thread who is calling close() can come from one of those who are calling normal methods. In that situation
 * this calling thread will wait forever as it is waiting itself to return. 
 * 
 * This class keeps a per-thread counter to avoid the above deadlock. Before waiting, the thread can use this class 
 * to exempt itself from the counter, thus eliminating the possibility of hang on itself.
 *
 * @author <a href="mailto:hgao@redhat.com">Howard Gao</a>
 * 
 * Created Dec 9, 2009 4:37:14 PM
 *
 *
 */
public class ThreadValveCounter
{

   private ConcurrentHashMap<Long, IntHolder> threadCounter = new ConcurrentHashMap<Long, IntHolder>();

   public void pushThread()
   {
      Long tid = Thread.currentThread().getId();
      IntHolder counter = threadCounter.get(tid);
      if (counter == null)
      {
         counter = new IntHolder();
         threadCounter.put(tid, counter);
      }
      counter.add();
   }

   public int popThread()
   {
      IntHolder counter = threadCounter.get(Thread.currentThread().getId());
      if (counter != null)
      {
         if (counter.release() == 0)
         {
            threadCounter.remove(Thread.currentThread().getId());
            synchronized (threadCounter)
            {
               threadCounter.notifyAll();
            }
         }
      }
      return 0;
   }

   public void waitForCompletion()
   {
      exemptCurrentThread();
      synchronized (threadCounter)
      {
         while (getCallCount() > 0)
         {
            try
            {
               threadCounter.wait();
            }
            catch (InterruptedException e)
            {
               // ignore
            }
         }
      }
   }

   private void exemptCurrentThread()
   {
      IntHolder counter = threadCounter.get(Thread.currentThread().getId());
      if (counter != null)
      {
         counter.clear();
      }
      synchronized (threadCounter)
      {
         threadCounter.notifyAll();
      }
   }

   private int getCallCount()
   {
      synchronized (threadCounter)
      {
         Iterator<IntHolder> iter = threadCounter.values().iterator();
         int total = 0;
         while (iter.hasNext())
         {
            total += iter.next().get();
         }
         return total;
      }
   }

   private static class IntHolder
   {
      private int counter = 0;

      public void add()
      {
         counter++;
      }

      public int get()
      {
         return counter;
      }

      public int release()
      {
         counter--;
         return counter;
      }

      public void clear()
      {
         counter = 0;
      }
   }

}
